"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.discoverSignatureComment = exports.discoverComment = void 0;
const ts = require("typescript");
const models_1 = require("../../models");
const paths_1 = require("../../utils/paths");
// Note: This does NOT include JSDoc syntax kinds. This is important!
// Comments from @typedef and @callback tags are handled specially by
// the JSDoc converter because we only want part of the comment when
// getting them.
const wantedKinds = {
    [models_1.ReflectionKind.Project]: [ts.SyntaxKind.SourceFile],
    [models_1.ReflectionKind.Module]: [ts.SyntaxKind.SourceFile],
    [models_1.ReflectionKind.Namespace]: [
        ts.SyntaxKind.ModuleDeclaration,
        ts.SyntaxKind.SourceFile,
        ts.SyntaxKind.BindingElement,
        ts.SyntaxKind.ExportSpecifier,
    ],
    [models_1.ReflectionKind.Enum]: [
        ts.SyntaxKind.EnumDeclaration,
        ts.SyntaxKind.VariableDeclaration,
    ],
    [models_1.ReflectionKind.EnumMember]: [
        ts.SyntaxKind.EnumMember,
        // These here so that @enum gets comments
        ts.SyntaxKind.PropertyAssignment,
        ts.SyntaxKind.PropertySignature,
    ],
    [models_1.ReflectionKind.Variable]: [
        ts.SyntaxKind.VariableDeclaration,
        ts.SyntaxKind.BindingElement,
        ts.SyntaxKind.ExportAssignment,
    ],
    [models_1.ReflectionKind.Function]: [
        ts.SyntaxKind.FunctionDeclaration,
        ts.SyntaxKind.BindingElement,
    ],
    [models_1.ReflectionKind.Class]: [
        ts.SyntaxKind.ClassDeclaration,
        ts.SyntaxKind.BindingElement,
    ],
    [models_1.ReflectionKind.Interface]: [ts.SyntaxKind.InterfaceDeclaration],
    [models_1.ReflectionKind.Constructor]: [ts.SyntaxKind.Constructor],
    [models_1.ReflectionKind.Property]: [
        ts.SyntaxKind.PropertyDeclaration,
        ts.SyntaxKind.PropertySignature,
        ts.SyntaxKind.BinaryExpression,
        ts.SyntaxKind.PropertyAssignment,
    ],
    [models_1.ReflectionKind.Method]: [
        ts.SyntaxKind.FunctionDeclaration,
        ts.SyntaxKind.MethodDeclaration,
    ],
    [models_1.ReflectionKind.CallSignature]: [
        ts.SyntaxKind.FunctionDeclaration,
        ts.SyntaxKind.VariableDeclaration,
        ts.SyntaxKind.MethodDeclaration,
        ts.SyntaxKind.MethodDeclaration,
        ts.SyntaxKind.PropertyDeclaration,
        ts.SyntaxKind.PropertySignature,
        ts.SyntaxKind.CallSignature,
    ],
    [models_1.ReflectionKind.IndexSignature]: [ts.SyntaxKind.IndexSignature],
    [models_1.ReflectionKind.ConstructorSignature]: [ts.SyntaxKind.ConstructSignature],
    [models_1.ReflectionKind.Parameter]: [ts.SyntaxKind.Parameter],
    [models_1.ReflectionKind.TypeLiteral]: [ts.SyntaxKind.TypeLiteral],
    [models_1.ReflectionKind.TypeParameter]: [ts.SyntaxKind.TypeParameter],
    [models_1.ReflectionKind.Accessor]: [
        ts.SyntaxKind.GetAccessor,
        ts.SyntaxKind.SetAccessor,
    ],
    [models_1.ReflectionKind.GetSignature]: [ts.SyntaxKind.GetAccessor],
    [models_1.ReflectionKind.SetSignature]: [ts.SyntaxKind.SetAccessor],
    [models_1.ReflectionKind.ObjectLiteral]: [ts.SyntaxKind.ObjectLiteralExpression],
    [models_1.ReflectionKind.TypeAlias]: [ts.SyntaxKind.TypeAliasDeclaration],
    [models_1.ReflectionKind.Reference]: [
        ts.SyntaxKind.NamespaceExport,
        ts.SyntaxKind.ExportSpecifier,
    ],
};
function discoverComment(symbol, kind, logger) {
    // For a module comment, we want the first one defined in the file,
    // not the last one, since that will apply to the import or declaration.
    const reverse = symbol.declarations?.some(ts.isSourceFile);
    const discovered = [];
    for (const decl of symbol.declarations || []) {
        const text = decl.getSourceFile().text;
        if (wantedKinds[kind].includes(decl.kind)) {
            const node = declarationToCommentNode(decl);
            if (!node) {
                continue;
            }
            // Special behavior here! We temporarily put the implementation comment
            // on the reflection which contains all the signatures. This lets us pull
            // the comment on the implementation if some signature does not have a comment.
            // However, we don't want to skip the node if it is a reference to something.
            // See the gh1770 test for an example.
            if (kind & models_1.ReflectionKind.ContainsCallSignatures &&
                !node.body &&
                node.kind !== ts.SyntaxKind.BindingElement) {
                continue;
            }
            const comments = ts.getLeadingCommentRanges(text, node.pos);
            if (reverse) {
                comments?.reverse();
            }
            const lastDocComment = comments?.find((c) => text[c.pos] === "/" &&
                text[c.pos + 1] === "*" &&
                text[c.pos + 2] === "*");
            if (lastDocComment) {
                discovered.push([decl.getSourceFile(), lastDocComment]);
            }
        }
    }
    switch (discovered.length) {
        case 0:
            return undefined;
        case 1:
            return discovered[0];
        default: {
            logger.warn(`${symbol.name} has multiple declarations with a comment. An arbitrary comment will be used.`);
            const locations = discovered.map(([sf, { pos }]) => {
                const path = (0, paths_1.nicePath)(sf.fileName);
                const line = ts.getLineAndCharacterOfPosition(sf, pos).line + 1;
                return `${path}:${line}`;
            });
            logger.info(`The comments for ${symbol.name} are declared at:\n\t${locations.join("\n\t")}`);
            return discovered[0];
        }
    }
}
exports.discoverComment = discoverComment;
function discoverSignatureComment(declaration) {
    const node = declarationToCommentNode(declaration);
    if (!node) {
        return;
    }
    const text = node.getSourceFile().text;
    const comments = ts.getLeadingCommentRanges(text, node.pos);
    const comment = comments?.find((c) => text[c.pos] === "/" &&
        text[c.pos + 1] === "*" &&
        text[c.pos + 2] === "*");
    if (comment) {
        return [node.getSourceFile(), comment];
    }
}
exports.discoverSignatureComment = discoverSignatureComment;
/**
 * Check whether the given module declaration is the topmost.
 *
 * This function returns TRUE if there is no trailing module defined, in
 * the following example this would be the case only for module <code>C</code>.
 *
 * ```
 * module A.B.C { }
 * ```
 *
 * @param node  The module definition that should be tested.
 * @return TRUE if the given node is the topmost module declaration, FALSE otherwise.
 */
function isTopmostModuleDeclaration(node) {
    return node.getChildren().some(ts.isModuleBlock);
}
/**
 * Return the root module declaration of the given module declaration.
 *
 * In the following example this function would always return module
 * <code>A</code> no matter which of the modules was passed in.
 *
 * ```
 * module A.B.C { }
 * ```
 */
function getRootModuleDeclaration(node) {
    while (node.parent &&
        node.parent.kind === ts.SyntaxKind.ModuleDeclaration) {
        const parent = node.parent;
        if (node.name.pos === parent.name.end + 1) {
            node = parent;
        }
        else {
            break;
        }
    }
    return node;
}
function declarationToCommentNode(node) {
    if (node.parent?.kind === ts.SyntaxKind.VariableDeclarationList) {
        return node.parent.parent;
    }
    if (node.kind === ts.SyntaxKind.ModuleDeclaration) {
        if (!isTopmostModuleDeclaration(node)) {
            return;
        }
        else {
            return getRootModuleDeclaration(node);
        }
    }
    if (node.kind === ts.SyntaxKind.ExportSpecifier) {
        return node.parent.parent;
    }
    if ([
        ts.SyntaxKind.NamespaceExport,
        ts.SyntaxKind.FunctionExpression,
        ts.SyntaxKind.FunctionType,
        ts.SyntaxKind.FunctionType,
        ts.SyntaxKind.ArrowFunction,
    ].includes(node.kind)) {
        return node.parent;
    }
    return node;
}
